/**
 * This thread is responsible for holding the TCP connection to the GNet node.
 * @file gproxy/BuildThread.java
 **/
package gproxy;

import java.net.*;
import java.io.*;

/**
 * This thread is responsible for holding the TCP connection to the GNet node.
 * This includes sending queries to GNet and handling incoming content from GNet.
 * Sending requests is done by the code of this object, but usually from other
 * threads.
 * @author Krista Bennett
 * @author Christian Grothoff
 **/
public class BuildThread extends Thread {
 
    /**
     * This socket is connected to the GNet node
     **/
    private Socket socket_;
    
    /**
     * This is the stream to receive content from
     **/
    private BufferedInputStream bis_;
    
    /**
     * This is the stream to send requests to
     **/
    private BufferedOutputStream bos_;
    
    /**
     * Flag used to stop the Thread.
     **/
    private transient boolean terminate = false;

    /**
     * The listeners that we show incomming data to.
     **/
    Listener[] listeners_;
    
    /**
     * The current number of listeners.
     **/
    int listCount_;

    /**
     * Host to connect to.
     **/
    private final String hostname_;
    
    /**
     * Port
     **/
    private final char port_;

    /**
     * Initial number of listeners.
     **/
    private static final int INIT_LISTENERS = 32;
    
    /**
     * create the socket to send/receive
     * @param port port of the local GNet node
     * @param hostname name of the host running GNet
     **/
    public BuildThread(char port, String hostname) {
	this.hostname_ = hostname;
	this.port_ = port;
	listeners_ = new Listener[INIT_LISTENERS];
	listCount_ = 0;
    }

    /**
     * Receive content from GNet and show it to everyone who is listening
     **/
    public void run(){
	byte[] buffer 
	    = new byte[Constants.CONTENT_BLOCK_SIZE+Constants.HASH_SIZE];
	while (!terminate) {
	    while (bis_ == null) {
		reset();
		CFuncts.sleep(5);
	    }
	    int size = 0;
	    try {
		size = bis_.read(buffer,size,Constants.CONTENT_BLOCK_SIZE+Constants.HASH_SIZE-size);
	    } catch (Exception ioe) {
		size = 0;
	    }
	    while ((!terminate) && (size != buffer.length)) {
		System.err.println("Invalid reply from GNet (" + size + " bytes). Trying to revive socket."); 	
		CFuncts.sleep(5);
		reset();
		// retry
		try {
		    size = 0;
		    while (size < buffer.length) 
			size += bis_.read(buffer,size,Constants.CONTENT_BLOCK_SIZE+Constants.HASH_SIZE-size);
		} catch(Exception io){
		    size = 0; 
		}
	    } // loop until ok 
	    if (!terminate)
		new NotificationThread(buffer).start();
	}         
    }
    
    /**
     * Try to re-open the socket.
     **/
    private void reset() {
	try {
	    bos_.close();
	    bis_.close();
	    socket_.close();			
	} catch (java.lang.Exception ex) {
	    // this error does not count! (expected!)
	}
	try {
	    socket_=new Socket(hostname_, port_);
	    bis_ = new BufferedInputStream(socket_.getInputStream());
	    bos_ = new BufferedOutputStream(socket_.getOutputStream());
	} catch (java.lang.Exception ex) {
	    socket_ = null;
	    bis_ = null;
	    bos_ = null;
	}
    }


   
    /**
     * Add a listener.
     * @param l the listener 
     **/
    public void addListener(Listener l) {
	synchronized (listeners_) {
	    if (listeners_.length == listCount_) {
		Listener[] tmp = new Listener[listCount_<<1];
		System.arraycopy(listeners_,0,
				 tmp, 0,
				 listCount_);
		listeners_ = tmp;
	    }
	    listeners_[listCount_++] = l;
	}           
    }
    
    /**
     * Remove a listener.
     * @param l the listener 
     **/
    public void removeListener(Listener l) {
	synchronized (listeners_) {
	    for (int i=0;i<listCount_;i++)
                 if (listeners_[i] == l) {
                     listeners_[i] = listeners_[--listCount_];
		     return;
                 }
	}
    }  
	
    /**
     * Send a request to GNet. 
     * @param message the request that should be send (hashcode / filename)
     * @param priority the priority of the request...
     * @param ttl the time-to-live for the request
     **/
    public void sendData (byte[] message,
			  int priority,
			  int ttl) {
	//System.out.println("sending query " + 
	// LoopbackBuildThread.bytesToHex(message) + " with ttl " + ttl);
	// extra check
	if (bos_ == null) 
	    reset();
	if (bos_ == null) {
	    System.err.println("No connection to GNET! Query not send.");
	    return;
	}
	if (message.length != Constants.HASH_SIZE) 
	    throw new Error("invalid message: " + message);
	try {
	    byte[] out = new byte[4+4+Constants.HASH_SIZE];
	    Converter.setInt(out,Constants.PRIORITY_OFFSET,
			     Converter.reverseEndian(priority));
	    Converter.setInt(out,Constants.TTL_OFFSET,
			     Converter.reverseEndian(ttl));
	    System.arraycopy(message,0,
			     out,Constants.HASH_OFFSET,
			     Constants.HASH_SIZE);	    
	    synchronized (this) {
		bos_.write(out);
		bos_.flush();
	    }
	} catch (IOException io) {
	    System.err.println("Could not send request: " + io.toString());
	    reset();
	}
    }
    
    /**
     * Close the connection to the socket.
     **/
    public void close(){
	try {
	    synchronized(this) {
		if (terminate)
		    return; // nothing to do
		terminate = true;
		bos_.close();
		bis_.close();
		socket_.close();
	    }
	} catch(IOException iox) {
	    System.err.println("Error while closing the socket: " + iox.toString());
	} catch(NullPointerException npe) {
	    System.err.println("Error while closing the socket: " + npe.toString());
	}
    }
    
    /**
     * Thread that notifies each Listener for an
     * incoming packet.
     * @author Christian Grothoff
     **/
    final private class NotificationThread extends Thread {
	
	/**
	 * The incoming packet.
	 **/
	private final byte[] buffer_;
	
	/**
	 * Create the Thread.
	 * @param buffer the incoming packet
	 **/
	private NotificationThread (byte[] buffer) {
	    this.buffer_ = buffer;
	}

	/**
	 * The main method of the Thread. Calls
	 * each Listener and then exits.
	 **/
	public void run() {
	    // System.err.println("GProxy: receiving data"+listCount_);
	    byte[] data = new byte[Constants.CONTENT_BLOCK_SIZE];	
	    byte[] hash = new byte[Constants.HASH_SIZE];
	    System.arraycopy(buffer_,0,
			     hash,0,
			     Constants.HASH_SIZE);
	    System.arraycopy(buffer_,Constants.HASH_SIZE,
			     data,0,
			     Constants.CONTENT_BLOCK_SIZE);
	    //System.out.println("received data " + 
	    //		       LoopbackBuildThread.bytesToHex(data));
	    synchronized (listeners_) {
		for (int i=0;i<listCount_;i++) {
		    try {
			listeners_[i].receive(hash, data);
		    } catch (Throwable t) {
			if (GProxy.DEBUG) {
			    System.err.println("BuildThread received exception: " + t);
			    t.printStackTrace();
			}
		    }
		}
	    }
	}

    } // end of NotificationThread
    
} // end of BuildThread
