/**
 * QueryCoordinator repeats queries to gnet for a
 * given search query until the Search-Window is
 * closed by the user. It notifies the QueryDisplay
 * about new files that match the query.
 * @file gproxy/QueryCoordinator.java
 **/
package gproxy;

import java.util.Hashtable;

/**
 * QueryCoordinator repeats queries to gnet for a
 * given search query until the Search-Window is
 * closed by the user. It notifies the QueryDisplay
 * about new files that match the query.
 * @author Christian Grothoff
 * @author Krista Bennett
 **/
public class QueryCoordinator 
    extends Thread 
    implements Listener {

    /**
     * The QueryParser that determines what we
     * are looking for.
     **/
    private final QueryParser qp_;

    /**
     * BuildThread - the object asked for content.
     **/
    private final BuildThread bt_;

    /**
     * Root-Nodes received so far. Used to make
     * sure we don't display the same result
     * again and again. Maps data (1k byte[]) to
     * the Root-node.
     **/
    private final Hashtable roots_;

    /**
     * Current priorities of the queries
     **/
    private int[] priorities_;

    /**
     * Sum of the squares of the bytes in the
     * hashcodes of the queries. Used for an
     * approximate comparison. 
     **/
    private long[] doubleHashSum_;

    /**
     * This field is set to true if the QueryWindow
     * was closed and we thus can stop asking questions.
     **/
    private boolean done_;

    /**
     * Create a new QueryCoordinator. Creating this
     * object automatically starts the Thread.
     * @param qp QueryParser that determines what we
     *         are looking for.
     **/
    public QueryCoordinator(QueryParser qp,
			    BuildThread bt) {
	this.qp_ = qp;
	this.bt_ = bt;
	this.roots_ = new Hashtable();
	bt_.addListener(this);
	start(); // start this thread
    }

    /**
     * Stop asking questions, the user doesn't care
     * anymore...
     **/
    public void terminate() {
	if (done_)
	    throw new Error("QueryCoordinator terminated twice!?");
	done_ = true;
	bt_.removeListener(this);
    }

    /**
     * This method sends out the requests that
     * match the query.
     **/
    public void run() {
	String[] queryStrings = qp_.getQueries();
	priorities_ = new int[queryStrings.length];
	int ttl = 300;
	byte[][] doubleHash = new byte[queryStrings.length][];
	doubleHashSum_ = new long[queryStrings.length];
	for (int i=0;i<queryStrings.length;i++) {
	    priorities_[i] = qp_.getPriority();
	    doubleHash[i] = CFuncts.hash(CFuncts.hash(queryStrings[i].getBytes()));
	    for (int j=0;j<doubleHash[i].length;j++)
		doubleHashSum_[i]+=doubleHash[i][j]*doubleHash[i][j];
	}
	while (!done_) {
	    for (int i=0;i<queryStrings.length;i++) {
		bt_.sendData(doubleHash[i], 
			     priorities_[i],
			     ttl);
		priorities_[i]++;
	    }
	    ttl += 150;
	    CFuncts.sleep(150);
	}
    }

    /**
     * This method receives data corresponding to 
     * the indicated filename (hashcode)
     * @param hash hashcode representing filename
     * @param data data corresponding to this node
     **/ 
    public void receive(byte[] hash, byte[] data) {
	ByteArray ba = new ByteArray(data);
	if (roots_.get(ba) != null)
	    return; // seen before
	long l=0;
	for (int j=0;j<hash.length;j++)
	    l+=hash[j]*hash[j];
	for (int i=0;i<doubleHashSum_.length;i++)
	    if (doubleHashSum_[i] == l)
		priorities_[i] = 0;
	RootNode rn = new RootNode(hash, data);
 	String[] queryStrings = qp_.getQueries();
	for (int i=0;i<queryStrings.length;i++) 
	    if (rn.decrypt(queryStrings[i])) {
		if (GProxy.DEBUG)
		    System.out.println("QueryCoordinator: RootNode created: " + rn.toString());
		if (qp_.checkForDownload(rn)) 
		    roots_.put(data, rn);
		break;
	    }
    }

    

} // end of QueryCoordinator
