/**
 * Main function to insert files into gnet.
 * @author Christian Grothoff
 * @author Krista Bennett
 * @file gproxy/insertmain.c 
 **/

#include "insertfile.h"
#include "hostkey.h"
#include "lookup.h"
#include <unistd.h>
#include <getopt.h>
#include "cron.h"

/**
 * Prints the usage information for this command if the user errs.
 * Aborts the program.
 **/
void printhelp() {
  printf("USAGE: gnetinsert [OPTIONS] DESCRIPTION KEYWORDS*\n");
  printf("-h, --help                    : print this page\n");
  printf("-c FILENAME, --config=FILENAME: load config file (defaults: ~/.gnet)\n");
  printf("-v, --version                 : print the version number\n");
  printf("-f FILENAME, --file=FILENAME  : insert FILENAME (obligatory)\n");
  printf("-i, --index                   : index only, encode on-demand\n");
}

/**
 * the main function to insert files into GNet.
 * @param argc number of arguments from the command line
 * @param argv command line arguments
 * @return return value from insertFile: 0: ok, -1: file not found 
 **/   
int main(int argc, char ** argv) {
  int res;
  char* confFile;
  char* filename;
  char* description;
  char ** keywords;
  int num_keywords;
  int c;
  
  unsigned int encode_content;
  
  /* set default config file */
  confFile = "~/.gnet";
  /* assume we encode content unless otherwise specified */
  encode_content = 0; /* full encoding */
  filename = NULL;  
  while (1) {
    int option_index=0;
    static struct option long_options[] = {
      { "config",  1, 0, 'c' },
      { "version", 0, 0, 'v' },
      { "help",    0, 0, 'h' },
      { "index",   0, 0, 'i' },
      { "file",    1, 0, 'f' },
      { 0,0,0,0 }
    };    
    c = getopt_long(argc,
		    argv, 
		    "vhic:f:", 
		    long_options, 
		    &option_index);    
    if (c == -1) 
      break;  // No more flags to process
    switch(c) {
    case 'c': 
      confFile = optarg; 
      print("Configfile specified: %s.\n",confFile);
      break;
    case 'f': 
      filename = optarg; 
      break;
    case 'v': 
      print("Version v0.3.0 (beta), gnetinsert v1.1.0\n");
      return 0;
    case 'h': 
      printhelp(); 
      return 0;
    case 'i':
      encode_content = 1; /* on demand */
      break;
    default: 
      print("Unknown option %c. Aborting.\nUse --help to get a list of options.\n",
	    c);
      return -1;
    } /* end of parsing commandline */
  } /* while (1) */
  if (argc - optind < 2) {
    print("Not enough arguments.");
    printhelp();
    return -1;
  }
  readConfig(confFile);
  initRAND(); 
  
  description = argv[optind++];
  keywords = &(argv[optind]);
  num_keywords = argc - optind;
  
  if (encode_content == 1) {
    initCron();
    initLookup();
    encode_content = appendFilename(filename);
  }
  res = insertFile(filename,
		   description,
		   keywords,
		   num_keywords,
		   encode_content);
  
  if (res == 0) {
    if (encode_content == 1)
	printf("File %s successfully indexed.\n", filename);
    else
	printf("File %s successfully inserted.\n", filename);
  } else
    printf("Error inserting the file (%d).\n",res);
  return res;
}  

/* end of insertmain.c */
