/**
 * This tests the process of assembling the file back after it was broken into
 * pieces.
 * @author Ioana Patrascu
 * @author Krista Bennett
 * @file gproxy/test/testinsert.c
 **/

#include "insertfile.h"
#include "tree_utils.h"

#define T_DEBUG 1
#define DEBUG2 1

/** 
 * this is a procedure that, given a SECOND hash, creates
 * the filename and reads the file into memory and returns 
 * the CONTENT_BLOCK_SIZE bytes worth of data.
 *
 * @param char* filename_hash
 * @return char* data
 **/

char* getBlock(HashCode160* filename_hash) {
    char* filename;
    char* data;
    char* full_path;
    FILE* file;

    filename = hash2hex(filename_hash);

    if (T_DEBUG) printf("FROM GETBLOCK: %s\n", filename);

    full_path = makeContentFilename(filename);
    file=fopen(full_path,"r");
    free(full_path);

    if(file == NULL) {
        printf("\nBlock file %s not found\n", filename);
        exit(-1);
    }
    else 
        printf("File %s found.\n", filename);


    data = malloc(CONTENT_BLOCK_SIZE);

    if(fread(data, CONTENT_BLOCK_SIZE, 1, file)!= 1) {
         printf("Error reading file %s", full_path);
         fclose(file);
         exit(-1);
    }
    fclose(file);

    return data;
      
}

/**
 * This is a procedure that, given an inode and an offset, returns
 * the hashcode STRING (not just the hashcode).
 * YOU ARE RESPONSIBLE FOR DEALLOCATING THE STRING AFTER RETURN.
 *
 * @param char* block - data block containing hashcode
 * @param int offset - offset of hashcode (in bytes)
 * @return char* hash_string
 *
 **/
HashCode160* getHCodeFromBlock(char* block, int offset) {
   char* hash_string;
   HashCode160* hashcode = malloc(HASHED_BLOCK_SIZE);
   memcpy(hashcode, block + offset, HASHED_BLOCK_SIZE); 
   hash_string = hash2hex(hashcode);

   if (T_DEBUG) printf("ghcfb: %s\n", hash_string);

   free(hash_string);
   return hashcode;
}

/**
 * this function takes a block of data and an offset (filename first hash)
 * and returns the block of data associated with that file.
 *
 * @param char* block - data from which to pull hashcode
 * @param int offset - offset from which to read hashcode
 * @return return_block corresponding to decoded filename
 **/

char* getBlockFromParent(char* block, int offset) {
   char* first_hash_string;
   HashCode160* first_hash;
   HashCode160* second_hash;
   char* return_block;
   char* temps;

   first_hash = getHCodeFromBlock(block, offset);
   first_hash_string = hash2hex(first_hash);

   if (T_DEBUG) printf("1st(gbfp): %s\n", first_hash_string);

   second_hash = getSecondHash(first_hash);

   if (T_DEBUG) printf("**** 2nd(gbfp): %s\n", hash2hex(second_hash));

   return_block = getBlock(second_hash);

   free(first_hash_string);
   free(second_hash);
   return return_block;
}   

/**
 * this is a procedure that hashes the second time a keyword, gets the name 
 *  	of the lower level one and fhecks the fule to exist
 * @param block - data structure to read hashcode to get filename from
 * @param size - size of the block to be read from memory
 * @param pos - position in file from where it is supposed to read
 **/
unsigned char* hhelp(char* block, int size, int pos){
    HashCode160 hashedKeyword; 
    char* hash_of_hash;
    FILE* file; 
    char* full_path; /* full path of file from content dir */
    char* newBlock;

    newBlock= hash2hex((HashCode160*)block); 
    hashNBang(newBlock, strlen(newBlock),
	      &hashedKeyword);

    hash_of_hash = hash2hex(&hashedKeyword);

    if (T_DEBUG) fprintf(stderr, "testinsert: **2** hkw: %s dhkw: %s\n", newBlock, hash_of_hash);  

    full_path = makeContentFilename(hash_of_hash);
    file=fopen(full_path,"r");
    free(full_path);

    if(file == NULL) {
         printf("\nNode file %s not found\n", hash_of_hash);
         return (char*)(-1);
    }
    
    fseek(file, pos, SEEK_SET); 
    if(fread(newBlock, size, 1, file) != 1) {
         printf("Error reading file %s", file);
	 fclose(file);
	 return (char*)(-1);
    }
    fclose(file);
    return newBlock;
}

/**
 * this is a recursive function that traverses the tree and writes the leaves
 * 	to a file
 * @param block is the inode or data block
 * @param level is the current level in the tree
 * @param tree_level is the tree level used for comparison
 * @param tree_degree is the number of entires per inode
 * @param rfile file name where the procedure reconstructs the initial file
 *
 * Internal inodes look like this:
 * --------------------------------------------------------------
 * | CRC of| HC1 | HC2 | HC3 | HC4 |         .  .  .     | HC   |
 * | nodes |     |     |     |     |                     | tree | 
 * | below |     |     |     |     |                     | dgree|
 * --------------------------------------------------------------
 *    32     160   160   160   160                         160 <--- in bits
 *     4      20    20    20    20                          20 <--- in BYTES
 *           --------------- tree degree of these here ------
 **/ 
int unroll_Block(unsigned char* block, int level, int tree_level, 
                 int tree_degree, FILE* rfile, int size_processed,
                 int filesize) {
    int i; 
    char* block_string;
    char* child_block;

    if (T_DEBUG) printf("Level : %d - tree_level : %d\n", level, tree_level);
     
    for(i = 0; i < tree_degree && size_processed < filesize; i++) {
	if (level < tree_level){
            if (DEBUG2) fprintf(stderr, "In level < tree_level\n");
 	    level++;

            /* get the data corresponding to the hashcode at this */
            /* index of this node                                 */
            if (DEBUG2) printf("******************** level is %d, next child is %d ****************** \n\n", level, i);
            child_block = getBlockFromParent(block, CHECKSUM_SIZE +
                                               (i*HASHED_BLOCK_SIZE));
           
	    size_processed = unroll_Block(child_block, level, 
                tree_level, tree_degree, rfile, size_processed, filesize);	
	    free(child_block);
            level--;
	}
	else {
            if (DEBUG2) fprintf(stderr, "Not In level < tree_level\n");

            if ((filesize - size_processed) > CONTENT_BLOCK_SIZE) {
	       // write block to file
               fwrite(block, CONTENT_BLOCK_SIZE, 1, rfile);
               size_processed += CONTENT_BLOCK_SIZE;
            }
            else {
               // write last part of last block to file
               fwrite(block, filesize - size_processed, 1, rfile);
               size_processed += (filesize - size_processed);
            }

            /* Block is freed by parent call! */
            break;
	}

    }	//end for loop 

    return size_processed;
}


int main(int argc, char** argv) {
    struct stat st; 	/* file info */

    FILE* file;		/* file pointer for root node */
    FILE* rfile;	/* reconstructed file */

    char* keyword;	/* first hash of keyword */ 
    char* cmd_kword;	/* keyword from the command-line */
    char* cmd_fname;	/* filename of file to be inserted from cmd line */
    char* cmd_desc;	/* description to be used for file to be inserted */
    char* diff_string;	/* string used by system command to execute diff */
    char* h2_hkey; 	/* data to be hashed (???) */	
    char* hashkey;	/* hash of the keyword (???) */
    char* root;		/* root of the tree */
    char* root_string;  /* string representing the root filename */
    char* block; 	/* a block corresponding to an inode */
    char* full_path; 	/* full path of file from content dir */

    int32_t filesize;	/* filesize from root info node */

    int kword_len;	/* length of keyword from cmd lin */
    int sys_return;	/* return value from diff of reconstructed and orig. */

    HashCode160 hashedKeyword; /* hash of the keyword */ 
    struct datatree* tree = malloc(sizeof(struct datatree));

    if (argc != 4) {
       fprintf(stderr, "USAGE: testinsert <filename> <keyword> <description>\n");
       exit(-1);
    }

    cmd_fname = argv[1];
    cmd_kword = argv[2];
    cmd_desc = argv[3];

    kword_len = strlen(cmd_kword);

    insertFile(cmd_fname, cmd_kword, cmd_desc);

    /* use keywords to recover the files */

    if (T_DEBUG) printf("Starting the hashing...\n");
    hashNBang(cmd_kword,
	      kword_len,
	      &hashedKeyword);
    keyword = hash2hex(&hashedKeyword);
 
    if (T_DEBUG) printf("Will try to malloc\n");
    h2_hkey = malloc(BLOCK_LENGTH_SIZE+strlen(keyword));
 
    if (T_DEBUG) printf("Will try to memcpy\n");
    memcpy(h2_hkey + BLOCK_LENGTH_SIZE, keyword, strlen(keyword));

    if (T_DEBUG) printf("Will try to get the second hash\n");
    ((BLOCK_LENGTH*)(h2_hkey))[0] = strlen(keyword);

    hashedKeyword = hash(h2_hkey);
    free(h2_hkey);

    block = getBlock(hashedKeyword);

    /* get file size */ 

    filesize = ((int32_t*)(block))[0]; 
    if (T_DEBUG) printf("%d size of file \n", filesize);

    /* set up tree */

    treeSetup(tree, filesize);
    if (T_DEBUG) printTreeInfo(tree);

    /* CRC check */
 
    /* get hash of data */
    /* read the hash of the root inode */
    root = malloc(HASHED_BLOCK_SIZE);
    memcpy(root, block + LENGTH_SIZE + CHECKSUM_SIZE, HASHED_BLOCK_SIZE);


    free(block);

    /* Now that we have the root hash code, turn it into a string so */
    /* that we can 1. unencrypt it (eventually) and 2. hash it again */
    /* to get the filename.                                          */

    root_string = hash2hex((HashCode160*)root);

    if (T_DEBUG) printf("Root string is %s\n", root_string);

    h2_hkey = malloc(BLOCK_LENGTH_SIZE + strlen(root_string));
    memcpy(h2_hkey + BLOCK_LENGTH_SIZE, root_string, strlen(root_string));

    ((BLOCK_LENGTH*)(h2_hkey))[0] = (BLOCK_LENGTH)(strlen(root_string));
 
    hashedKeyword = hash(h2_hkey);
    free(h2_hkey);

    /* get root inode block */
    block = getBlock(hashedKeyword);

    rfile = fopen("reconstrFile","w");
    if(rfile == NULL) {
	printf("ERROR opening reconstruction file!!\n");
	return -1;
    } 

    unroll_Block(block, 1, tree->tree_height + 1, tree->tree_degree, 
                 rfile, 0, filesize); 

    /* clean up */
    free(block);
    fclose(rfile);

    /* Here is where the test really is - compare the reconstructed file */
    /* with the original.                                                */

    diff_string = malloc(strlen("diff -a reconstrFile ") + 
                         strlen(cmd_fname) + 1); 

    strcpy(diff_string, "diff -a reconstrFile ");
    strcat(diff_string, cmd_fname);

    sys_return = system(diff_string);
  
    return 0;
}

