/**
 * This file provides methods to manipulate host credit
 * rankings. It should be merged with outbuffer soon.
 * @author Tzvetan Horozov
 * @author Christian Grothoff
 **/ 

#ifndef CONNECTION_H
#define CONNECTION_H

#include "gnettypes.h"
#include "symcipher.h"
#include "sessionkey.h"
#include "identity.h"
#include "semaphore.h"
#include "checksum.h"
#include "query.h"
#include "configuration.h"
#include "policy.h"
#include "cron.h"

/* ********************* Methods called from "node" ********************** */

/**
 * Initialize this module.
 **/
void initConnection();


/* ********************** Send-interface ****************************** */

/**
 * Send a message to "many" people., except ourselves! TTH (enqueue)
 * @param message the message to send
 * @param priority how important is the message? The higher, the more important
 **/
void broadcast(MessagePart * message,
	       unsigned int priority,
	       unsigned int numberofhosts);

/**
 * Send a message to a specific host (reply, enqueue)
 * @param message the message to send (unencrypted!), first BLOCK_LENGTH_SIZE bytes give size
 * @param hostId the identity of the receiver
 * @param priority how important is the message?
 **/
void unicast(MessagePart * message,
	     HostIdentity * hostId,
	     unsigned int priority);


/**
 * We have gotten a query and decided to forward it
 * (or it may be our own query). This method must
 * decide to whom to send the query. 
 * <p>
 * Attention: if the returnTo-identity is not us, then
 * the receiver must also been given the identity of the
 * host we are forwarding (but not indirecting) for.
 * @param msg the message to forward.
 **/
void forwardQuery(QUERY_Message * msg);

/* *********************** Hostcredit interface ********************** */

/**
 * Increase the host credit by a value - synchronized
 * @param hostId is the identity of the host
 * @param value is the int value by which the host credit is to be increased
 * @returns the new credit
 */
unsigned int changeHostCredit(HostIdentity * hostId, int value);

/**
 * Obtain the credit record of the host.
 **/
unsigned int getHostCredit(HostIdentity * hostId);

/* ************************* encryption service ********************** */

/**
 * Prepare data to be send to the given host. Caller is
 * responsible for ensuring that a sessionkey exists 
 * @param data the data to encrypt
 * @param size how big is data?
 * @param hostId the target host the encryption is for
 * @param result where to store the encrypted result
 * @return -1 on error, otherwise size of result
 **/
int encryptForHost(void * data,
		   BLOCK_LENGTH size,
		   HostIdentity * hostId,
		   BLOWFISHEncryptedData * result);

/**
 * Decipher data comming in from a foreign host.
 * @param data the data to decrypt
 * @param size how big is data?
 * @param hostId the sender host that encrypted the data 
 * @param result where to store the decrypted data
 * @returns the size of the decrypted data, -1 on error
 **/
int decryptFromHost(BLOWFISHEncryptedData * data,
		    BLOCK_LENGTH size,
		    HostIdentity * hostId,
		    void * result);  

/* ************************** internal: I/O *********************** */


/* ********************* Key exchange and connection maintanance ********** */

/**
 * We received a sign of life from this host.
 **/
void notifyActive(HostIdentity * hostId);

/**
 * We received a sign of life from this host. 
 **/
void notifyActiveSKEY(HostIdentity * hostId,
		      HashCode160 * hashcode);

/**
 * Force adding of a host to the buffer.
 **/
BufferEntry * addHost(HostIdentity * hostId,
		      int doExchangeSkey);

/**
 * Force adding of a host to the buffer. 
 * @param created the time at which this key was created
 * @returns 1 on error, 0 on success
 **/
int addHostWithSKEY(HostIdentity * hostId,
		    SESSIONKEY * skey,
		    time_t created);

/**
 * Send a buffer
 **/
 void sendBufferOfHost(HostIdentity * hostId);

/* ******************** Internal: helpers ******************************** */
 
/**
 * Compute the hashtable index of a host id.
 **/
unsigned int computeIndex(HostIdentity * hostId);

/**
 * Append a message to the current buffer.
 **/
void appendToBuffer(BufferEntry * be,
		    MessagePart * message);

/**
 * Send a buffer
 **/
void sendBuffer(BufferEntry * be);



#endif
/* end of connection.h */

