/**
 * This file provides methods to manipulate host credit
 * rankings. It should be merged with outbuffer soon.
 * @author Tzvetan Horozov
 * @author Christian Grothoff
 **/ 

#include "connection.h"

/**
 * The buffer containing all current connections.
 **/
BufferEntry * CONNECTION_buffer_;

/**
 * Size of the CONNECTION_buffer_
 **/
unsigned int CONNECTION_MAX_HOSTS_;

/**
 * The DirContents array for scanning the hosts/ directory.
 **/
DirContents * CONNECTION_knownHosts;
int CONNECTION_currentMaxKnownHosts;
int CONNECTION_currentKnownHosts;
Semaphore * CONNECTION_knownHostsLock;

DirName * CONNECTION_networkIdDirectory;
DirName * CONNECTION_credibilityDirectory;

/* ************* internal Methods **************** */

/**
 * Call this method periodically to decrease liveness of hosts.
 * Call once every minute.
 **/
void cronDecreaseLiveness();

/**
 * Call this method periodically to scan data/hosts for new hosts.
 **/
void cronScanForHosts();

/**
 * Read host-information from a file.
 **/
void readHost(BufferEntry * be);

/**
 * Write host-infromation to a file - flush the buffer entry!
 **/
void writeHost(BufferEntry * be);

/**
 * Look in the list for known hosts; pick a random host
 * for the hosttable at index index.
 **/
void scanForHosts(unsigned int index,
		  DirContents * knownHosts);

/**
 * Look for a host in the table.
 * @returns the index of the host in the table
 **/
BufferEntry * lookForHost(HostIdentity * hostId);

/* ******************** CODE ********************* */


/**
 * Initialize this module.
 **/
void initConnection() {
  unsigned int i;
  BufferEntry * be;

  CONNECTION_MAX_HOSTS_ = getMaxNodes();
  i = 1;
  while (i<= CONNECTION_MAX_HOSTS_)
    i*=2;
  CONNECTION_MAX_HOSTS_ = i; /* make sure it's a power of 2 */

  CONNECTION_buffer_ 
    = (BufferEntry*) malloc(sizeof(BufferEntry)*CONNECTION_MAX_HOSTS_);

#ifdef PRINT_CONNECTION
  print("CONNECTION: Connection module initialization\n");
#endif
  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) {
    be = &CONNECTION_buffer_[i];
    be->sem  = new_semaphore(1);  
    be->credit = 0;
    be->isAlive = 0;
    be->prio = 0;
    be->buffer = malloc(BUFSIZE+BLOCK_LENGTH_SIZE);
    be->len = 0;
  }
  CONNECTION_networkIdDirectory
    = (DirName*)expandFileName((FileName)getNetworkIdDirectory());
  CONNECTION_credibilityDirectory
    = (DirName*)expandFileName((FileName)getCreditDirectory());
  
  CONNECTION_currentMaxKnownHosts 
    = CONNECTION_MAX_HOSTS_;
  CONNECTION_knownHosts 
    = malloc(sizeof(HexName)*CONNECTION_currentMaxKnownHosts);
  CONNECTION_currentKnownHosts 
    = scanDirectory(CONNECTION_networkIdDirectory,
		    CONNECTION_knownHosts,
		    CONNECTION_currentMaxKnownHosts);
  CONNECTION_knownHostsLock = new_semaphore(1);
#ifdef PRINT_CONNECTION
  print("CONNECTION: working with %d known hosts\n",
	CONNECTION_currentKnownHosts);
#endif
  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) 
    scanForHosts(i, CONNECTION_knownHosts);
  i = 0;
#ifdef PRINT_CONNECTION
  print("CONNECTION: Connection module initialization complete\n");
#endif

  /* set cron jobs */
  /* schedule cron job */
  addCronJob(&cronScanForHosts,
	     15*60, 15*60); /* scan every 15 minutes */  
  addCronJob(&cronDecreaseLiveness,
	     60, 60); /* Liveness in 1m intervals */  
}

/**
 * Print the contents of the connection buffer. May
 * NOT be called from synchronized context!
 **/
void printConnectionBuffer() {
  int i;
  BufferEntry * be;
  HexName hostName;
  SKEYString skey;

  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) {
    be = &CONNECTION_buffer_[i];
    semaphore_down(be->sem);
    if (be->isAlive > 0) {
      hash2hex(&be->hostId.hashPubKey, &hostName);
      printSessionKey(&be->skey, &skey);
      
      print("%d: %s has credit %d, liveness %u, buflen %d\n\t and key %s\n",
	    i, &hostName,
	    be->credit, be->isAlive, be->len,
	    &skey);
    }
    semaphore_up(be->sem);     
  }
}

/**
 * Call this method periodically to scan data/hosts for new hosts.
 **/
void cronScanForHosts() {
  semaphore_down(CONNECTION_knownHostsLock);   
  if (CONNECTION_currentKnownHosts == CONNECTION_currentMaxKnownHosts) {
    free(CONNECTION_knownHosts);
    CONNECTION_currentMaxKnownHosts += CONNECTION_MAX_HOSTS_;
    CONNECTION_knownHosts = malloc(sizeof(HexName)*CONNECTION_currentMaxKnownHosts);
  }
  CONNECTION_currentKnownHosts 
    = scanDirectory(CONNECTION_networkIdDirectory,
		    CONNECTION_knownHosts,
		    CONNECTION_currentMaxKnownHosts);
#ifdef PRINT_WARNINGS
   if (CONNECTION_currentKnownHosts == 0) 
    print("CONNECTION: scanDirectory returned no known hosts!\n");
#endif
   semaphore_up(CONNECTION_knownHostsLock);   
}

/**
 * Call this method periodically to decrease liveness of hosts.
 * Call once every minute.
 **/
void cronDecreaseLiveness() {
  int i;
  BufferEntry * be = NULL;
  unsigned int act;

#ifdef PRINT_CONNECTION
  print("CONNECTION: Connection module cron start with %d hosts\n",
	CONNECTION_currentKnownHosts);
  printConnectionBuffer();
#endif
  for (i=0;i<CONNECTION_MAX_HOSTS_;i++) {
    be = &CONNECTION_buffer_[i];
    semaphore_down(be->sem);
    act = be->isAlive;
    be->isAlive = act >> 1;
    if (act > 0) 
      writeHost(be);
    if (act == 0) {
      semaphore_down(CONNECTION_knownHostsLock);   
      scanForHosts(i, CONNECTION_knownHosts);           
      semaphore_up(CONNECTION_knownHostsLock);   
    }
    semaphore_up(be->sem);
  }
}

/**
 * We received a sign of life from this host. 
 **/
void notifyActiveSKEY(HostIdentity * hostId,
		      HashCode160 * hashcode) {
  BufferEntry * be;
  HashCode160 expected;
#ifdef PRINT_CONNECTION
  HexName hex;
#endif

  be = lookForHost(hostId);
  if (be == NULL)
    return;
  hash(be->skey.key,SESSIONKEY_LEN,
       &expected);
  if (!equalsHashCode160(hashcode, &expected)) {
    semaphore_up(be->sem);  
    return;
  }

#ifdef PRINT_CONNECTION
  hash2hex(&hostId->hashPubKey, &hex);
  print("CONNECTION: Marking host %s active.\n",
	&hex);
#endif
  be->isAlive |= 0x80000000; 
  semaphore_up(be->sem);  
#ifdef PRINT_CONNECTION
  print("CONNECTION: Done marking host active\n");
#endif
}

/**
 * We received a sign of life from this host. 
 **/
void notifyActive(HostIdentity * hostId) {
  BufferEntry * be;

#ifdef PRINT_CONNECTION
  HexName hex;

  hash2hex(&hostId->hashPubKey, &hex);
  print("CONNECTION: Marking host %s active.\n",
	&hex);
#endif
  be = lookForHost(hostId);
  if (be != NULL){
    be->isAlive |= 0x80000000; /* no locking: not much damage can be done */
    semaphore_up(be->sem);
  }
#ifdef PRINT_CONNECTION_ALL
  else
    print("CONNECTION: host not in current table.\n");
#endif
#ifdef PRINT_CONNECTION
  print("CONNECTION: Done marking host active\n");
#endif
}

/**
 * Send a message to "many" people., except ourselves! TTH (enqueue)
 * @param message the message to send
 * @param priority how important is the message? The higher, the more important
 **/
void broadcast(MessagePart * message,
	       unsigned int priority,
	       unsigned int numberofhosts) {
  unsigned int i;
  BufferEntry * be = NULL;

#ifdef PRINT_CONNECTION
  print("CONNECTION: broadcast to %d hosts\n", numberofhosts);
#endif
  i = 0;
  while ((i<CONNECTION_MAX_HOSTS_) && (numberofhosts > 0)) {
    /* we need no sync here as we only read,
       and concurrent rw access does not hurt */
    be = &CONNECTION_buffer_[i++];
    if (be->isAlive <= 0) 
      continue;    
    unicast(message,
	    &be->hostId,
	    priority);
    numberofhosts--;
  }
#ifdef PRINT_CONNECTION
  print("CONNECTION: Done with broadcast, %d packets not sent.\n",
	numberofhosts);
#endif
}

/**
 * We have gotten a query and decided to forward it
 * (or it may be our own query). This method must
 * decide to whom to send the query. 
 * <p>
 * Attention: if the returnTo-identity is not us, then
 * the receiver must also been given the identity of the
 * host we are forwarding (but not indirecting) for.
 * @param msg the message to forward.
 **/
void forwardQuery(QUERY_Message * msg) {
  int m;
#ifdef PRINT_ROUTING
  HexName hex;
#endif

  if (msg->ttl <= 0)
    return;
  m = CONNECTION_currentKnownHosts; /* the obvious max */
  if (m > CONNECTION_MAX_HOSTS_)
    m = CONNECTION_MAX_HOSTS_; /* the other obvious max */
  if ((msg->priority <= 1)&&(m>4)) 
    m = 4; /* if priority is REALLY low, send only to 4 hosts */
  else
    if (m > msg->priority)
      m = msg->priority; /* make sure priority does not
			    drop under 1 if possible */
  msg->priority = msg->priority/m;
  
#ifdef PRINT_ROUTING
  hash2hex(&msg->query, &hex);
  print("ROUTING: Forwarding query %s (broadcast) with ttl %d.\n",
	&hex,msg->ttl);
#endif
  broadcast(&msg->header,
	    msg->priority,
	    m);
}

/**
 * Send a message to a specific host (reply, enqueue)
 * @param message the message to send (unencrypted!), first BLOCK_LENGTH_SIZE bytes give size
 * @param hostId the identity of the receiver
 * @param priority how important is the message?
 **/
void unicast(MessagePart * message,
	     HostIdentity * hostId,
	     unsigned int priority) {
  BufferEntry * be;
#ifdef PRINT_CONNECTION
  HexName hex;
  hash2hex(&hostId->hashPubKey, &hex);
  print("CONNECTION: unicasting to host %s\n",
	&hex);
#endif  
  be = lookForHost(hostId);
  if (be == NULL) 
    be = addHost(hostId, 1);  
  if (be == NULL) {
#ifdef PRINT_WARNINGS
    print("unicast failed, host unknown!\n");
#endif
    return; /* error */
  }  
  if (be->prio < priority)
    be->prio = priority;
  appendToBuffer(be,
		 message);
#ifdef DUMP_PACKETS_SEND
  print("Unicast appended to buffer:\n");
  dumpPacket(be->len,be->buffer);
#endif
  semaphore_up(be->sem);
}

/**
 * Increase the host credit by a value. No point in synchronization
 * as changes can't mess up much anyway.
 * @param hostId is the identity of the host
 * @param value is the int value by which the host credit is to be increased
 * @returns the actual change in credit
 */
unsigned int changeHostCredit(HostIdentity * hostId, int value){
  BufferEntry * be;

#ifdef PRINT_CONNECTION
  HexName hex;
  hash2hex(&hostId->hashPubKey, &hex);
  print("CONNECTION: changing credit for host %s by %d\n",
	&hex, value);
#endif  
  be = lookForHost(hostId);
  if (be == NULL) {
    be = addHost(hostId,1);
    if (be == NULL)
      return 0;
  }
  if (be->credit + value < 0)
    be->credit = 0;
  else
    be->credit += value;
  semaphore_up(be->sem);
  return value;
}

/**
 * Obtain the credit record of the host.
 **/
unsigned int getHostCredit(HostIdentity * hostId) {
#ifdef PRINT_CONNECTION
  HexName hex;
  hash2hex(&hostId->hashPubKey, &hex);
  print("CONNECTION: looking for credit for host %s\n",
	&hex);
#endif
  return changeHostCredit(hostId,0);
}
			       
/**
 * Decipher data comming in from a foreign host. As usual,
 * no synchronization, as a concurrent access to the same
 * table entry would discard the sessionkey and thus make
 * decryption impossible anyway.
 * @param data the data to decrypt
 * @param hostId the sender host that encrypted the data 
 * @param result where to store the decrypted data, must be at least of size data->len long
 * @returns the size of the decrypted data, -1 on error
 **/
int decryptFromHost(BLOWFISHEncryptedData * data,
		    BLOCK_LENGTH size,
		    HostIdentity * hostId,
		    void * result) {  
  BufferEntry * be;
  int res;
#ifdef PRINT_CONNECTION
  HexName hex;
#endif
#ifdef PRINT_WARNINGS
#ifndef PRINT_CONNECTION
  HexName hex;
#endif
#endif
  if ((data == NULL) || (hostId == NULL)) {
    print("CONNECTION: could not decrypt, message or hostId was NULL\n");
    return -1;
  }

#ifdef PRINT_CONNECTION
  hash2hex(&hostId->hashPubKey, &hex);
  print("CONNECTION: decrypting message from host %s\n",
	&hex);
#endif
  be = lookForHost(hostId);
  if (be == NULL) {
#ifdef PRINT_WARNINGS
#ifndef PRINT_CONNECTION
    hash2hex(&hostId->hashPubKey, &hex);
#endif
    print("CONNECTION: decrypting message from host %s failed, no sessionkey!\n",
	  &hex);
#endif
    return -1; /* could not decrypt */
  }
  res = decryptBlock(&be->skey, 
		     data,
		     size,
		     result);
  semaphore_up(be->sem);
#ifdef PRINT_SKEY
  print("Decrypting message. Connection table is:\n");
  printConnectionBuffer();
#endif
  return res;
}

/* ********************* internal methods ************************ */

/**
 * Read host-information from a file. The semaphore
 * of the given be must already be down.
 **/
void readHost(BufferEntry * be) {
  DirName * dir;
  HexName fil;
  FileName fn;

#ifdef PRINT_CONNECTION
  HexName hostName;

  hash2hex(&be->hostId.hashPubKey, &hostName);
  print("CONNECTION: reading credit rating for host %s\n",
	&hostName);
#endif
  dir=getCreditDirectory();
  hash2hex(&be->hostId.hashPubKey,&fil);
  fn = malloc(strlen((char*)dir)+sizeof(HexName)+1);
  buildFileName(dir,&fil,fn); 
  if (sizeof(unsigned int) !=
      readFile(fn, sizeof(unsigned int), &be->credit))
    be->credit = 0;
  free(fn); 
}

/**
 * Perform a session key exchange for entry be.
 * First sends a HELO and then the new SKEY
 * (in one unencrypted packet). When called, the 
 * semaphore of at the given index must already be down
 **/
void exchangeKey(BufferEntry * be) {
  HELOSKEY m;
  int crc;

  /* sleep a random amount of time. This reduces the chance that 2 hosts
     perform a synchronous session key exchange (which is otherwise likely,
     as one host sends HELO and then the other eventually wants to 
     reply "immediately". If each host is forced to delay any action by
     a few seconds, the chance of this collision is reduced. */
  /*  sleep(rand() & 31); DID NOT HELP :-( */

#ifdef PRINT_SKEY
  HexName hex;
  hash2hex(&be->hostId.hashPubKey, &hex);
  print("SKEY: Beginning of exchange key with %s\n",
	&hex);
#endif

#ifdef PRINT_CONNECTION_ALL
  print("CONNECTION: EXCHANGING SKEY (incl. HELO exchange)\n");
#endif
  makeSessionkey(&be->skey);
  time(&be->created); /* set creation time for session key! */
  if (!makeSessionKeySigned(&be->hostId,
			    &be->skey,
			    be->created,
			    &m.sks))
    return;
  be->isAlive = 0; /* mark as dead until we get the receit! */
  /* send tmp to target */
  createHELO(&m.helo);

  /* get crc of data */
  crc = crc32N(&m, sizeof(HELOSKEY));
#ifdef DUMP_PACKETS_SEND 
  print("Exchanging key - sending in plaintext to host:\n");
  print("m.helo->size = %d (expected: %d)\n",
	((MessagePart*) &m.helo)->size,
	sizeof(HELO_Message)-BLOCK_LENGTH_SIZE);
  print("m.helo->size via message = %d\n",
	((MessagePart*)&(((Message)&m)[BLOCK_LENGTH_SIZE]))->size);
  print("m.skey->size %d == %d == %d?\n",
	((MessagePart*) &m.sks)->size,
	((MessagePart*)&(((Message)&m)[2*BLOCK_LENGTH_SIZE+((MessagePart*) &m.helo)->size]))->size,
	m.sks.header.size);
	
  dumpPacket((Message)&m);
#endif 
  sendToHost(&be->hostId,
	     sizeof(HELOSKEY),
	     &m, 
	     0, 
	     crc);
#ifdef PRINT_CONNECTION_ALL
  print("CONNECTION: done exchanging session key\n");
#endif    
#ifdef PRINT_SKEY
  print("End of exchange key with %s\n",
	&hex);
#endif
}

/**
 * Write host-infromation to a file - flush the buffer entry!
 * The semaphore must already be down when calling.
 **/
void writeHost(BufferEntry * be) {
  HexName fil;
  FileName fn;

#ifdef PRINT_CONNECTION
  print("CONNECTION: writing creditiblity to file\n");
#endif    

  if (be->isAlive == 0)
    return;
  hash2hex(&be->hostId.hashPubKey,&fil);
  fn = malloc(strlen((char*)CONNECTION_credibilityDirectory)+sizeof(HexName)+1);
  buildFileName(CONNECTION_credibilityDirectory,&fil,fn);
  writeFile(fn, &be->credit, sizeof(unsigned int));
  sendBuffer(be);
  free(fn);
#ifdef PRINT_CONNECTION
  print("CONNECTION: writing creditiblity to file done\n");
#endif 
}

/**
 * Look in the list for known hosts; pick a random host
 * for the hosttable at index index. When called, the 
 * semaphore of at the given index must already be down.
 **/
void scanForHosts(unsigned int index,
		  DirContents * CONNECTION_knownHosts) {
  int i;
  HostIdentity hi;
  BufferEntry * be;
  int matchCount = 0;

  be = &CONNECTION_buffer_[index];
  /* scan for known hosts, add them to the broadcast list */
  for (i=0;i<CONNECTION_currentKnownHosts;i++) {
    hex2hostIdentity(&CONNECTION_knownHosts->files[i],&hi);
    if (computeIndex(&hi) == index)
      matchCount++;
  }
  if (matchCount == 0)
    matchCount = -1; /* no match */
  else 
    matchCount = random() % matchCount; /* randomly choose host */
  for (i=0;i<CONNECTION_currentKnownHosts;i++) {
    hex2hostIdentity(&CONNECTION_knownHosts->files[i],&hi);
    if (computeIndex(&hi) == index) {
      if (matchCount == 0) {
    	memcpy(&be->hostId,
    	       &hi,
    	       sizeof(HostIdentity));
	readHost(be);
	exchangeKey(be);
	break;
      } else
	matchCount--; /* no, matchCount more entries to go */
    }
  }
#ifdef PRINT_CONNECTION_ALL
  print("CONNECTION: scanning directory for known hosts done\n");
#endif 
}

/**
 * Look for a host in the table. If the entry WAS there
 * at the time of checking, returns the Entry. Mind that
 * a concurrent access may make the return-value invalid.
 * @returns the index of the host in the table
 **/
BufferEntry * lookForHost(HostIdentity * hostId) {
  unsigned int hash;
  BufferEntry * be;

  hash = computeIndex(hostId);
  be = &CONNECTION_buffer_[hash];
  semaphore_down(be->sem);
  if (equalsHashCode160(&hostId->hashPubKey,
			&be->hostId.hashPubKey)) 
    return be;
  else {
    semaphore_up(be->sem);
    return NULL; /* not found! */
  }
}

/**
 * Force adding of a host to the buffer. Synchronizes on
 * the adding process, yet releases the lock afterwards.
 * Thus a concurrent access may change the host immediately
 * after it was added.
 **/
BufferEntry * addHost(HostIdentity * hostId,
		      int doExchangeKey) {
  BufferEntry * be;
  /* flush old entry */

#ifdef PRINT_CONNECTION
  HexName hex;

  hash2hex(&hostId->hashPubKey, &hex);
  print("CONNECTION: adding host %s to the table.\n",
	&hex);
#endif 
#ifdef PRINT_WARNINGS
#ifndef PRINT_CONNECTION
  HexName hex;
#endif
#endif

  be = lookForHost(hostId);
  if (be == NULL) {
#ifdef PRINT_WARNINGS
    hash2hex(&hostId->hashPubKey,&hex);
    print("addHost: lookForHost for %s returned NULL!\n",
	  &hex);
#endif
    return NULL;
  }
  if (be->isAlive > 0) 
    writeHost(be);
  /* add new entry */
  memcpy(&be->hostId,
	 hostId,
	 sizeof(HostIdentity));	 
  readHost(be);
  if (doExchangeKey) 
    exchangeKey(be);
  return be;
}

/**
 * Force adding of a host to the buffer.
 **/
int addHostWithSKEY(HostIdentity * hostId,
		    SESSIONKEY * skey,
		    time_t created) {
  BufferEntry * be;
#ifdef PRINT_SKEY
  HexName hex;
#endif
#ifdef PRINT_CONNECTION
#ifndef PRINT_SKEY
  HexName hex;
#endif
#endif
    
  be = lookForHost(hostId);
  if (be == NULL) {
#ifdef PRINT_SKEY
    hash2hex(&hostId->hashPubKey, &hex);
    print("Session key exchange denied, host %s unknown!\n",
	  &hex);
#endif
    return 1;
  }
  /* here we should *also* consider the trust-levels in the future! */
  if (!hostIdentityEquals(&be->hostId,
			  hostId) &&
      (be->isAlive > 0) &&
      (be->created > created)) {
#ifdef PRINT_SKEY
    print("Session key exchange denied, another host is in the table and active!\n");
#endif
    return 1;
  }
#ifdef PRINT_CONNECTION
  hash2hex(&hostId->hashPubKey, &hex);
  print("CONNECTION: adding host %s with sessionkey to table.\n",
	&hex);
#endif 
  memcpy(&be->hostId,
	 hostId,
	 sizeof(HostIdentity));	 
  memcpy(&be->skey,
	 skey,
	 SESSIONKEY_LEN);
  be->created = created;
#ifdef PRINT_CONNECTION_ALL
  print("CONNECTION: adding host %s with sessionkey to table done.\n",
	&hex);
#endif 
  semaphore_up(be->sem);
  return 0;
}

/**
 * Compute the hashtable index of a host id.
 **/
unsigned int computeIndex(HostIdentity * hostId) {
  unsigned int res = (((unsigned int)hostId->hashPubKey.a) & 
		      ((unsigned int)(CONNECTION_MAX_HOSTS_ - 1)));
#ifdef EXTRA_CHECKS
  if (res >= CONNECTION_MAX_HOSTS_)
    errexit("FATAL: CONNECTION_MAX_HOSTS_ not power of 2? (%d)\n",
	    CONNECTION_MAX_HOSTS_);
#endif
  return res;
}


/**
 * Append a message to the current buffer. This method
 * assumes that the access to be is already synchronized.
 **/
void appendToBuffer(BufferEntry * be,
		    MessagePart * message) {
  BLOCK_LENGTH exSize;
  BLOCK_LENGTH newSize;
  unsigned char * buffer;

#ifdef PRINT_CONNECTION
  HexName hex;

  hash2hex(&be->hostId.hashPubKey, &hex);
  print("CONNECTION: adding message of size %d to buffer of host %s.\n",
	message->size,
	&hex);
#endif 
  buffer = be->buffer;
  if (message == NULL) {
#ifdef PRINT_WARNINGS
    print("appendToBuffer got garbage. Ignored.\n");
#endif
    return;
  }
  
  newSize = message->size+BLOCK_LENGTH_SIZE;
  exSize = be->len;
  if (exSize + newSize > BUFSIZE) {
    if (newSize > BUFSIZE) {
#ifdef PRINT_WARNINGS      
      print("appendToBuffer fatal error: exSize: %d newsize: %d BUFSIZE %d\n",
	    exSize, newSize, BUFSIZE);
#endif
      return;
    }
    sendBuffer(be); /* clean up*/
    exSize = 0;
  }
  memcpy(&buffer[exSize], 	 
	 message, newSize); 
  be->len = exSize + newSize;
  if (exSize + newSize > THRESHOLD) 
    sendBuffer(be); 
#ifdef PRINT_CONNECTION
  print("CONNECTION: done appending message.\n");
#endif 
}


/**
 * Send the buffer that belongs to this host.
 **/
void sendBufferOfHost(HostIdentity * hostId){
  BufferEntry * be;

  be = lookForHost(hostId);
  if (be == NULL)
    return; 
  sendBuffer(be);
  semaphore_up(be->sem);
}


/**
 * Send a buffer; assumes that access to be is already 
 * synchronized.
 **/
void sendBuffer(BufferEntry * be){
  int crc;
  BLOWFISHEncryptedDataMax content;

#ifdef PRINT_CONNECTION
  HexName hex;
  
  hash2hex(&be->hostId.hashPubKey, &hex);
  print("CONNECTION: sending buffer of host %s.\n",
	&hex);
#endif 

#ifdef DUMP_PACKETS_SEND
#ifndef PRINT_CONNECTION
  HexName hex;
  
  hash2hex(&be->hostId.hashPubKey, &hex);
#endif
#endif
  if ((be == NULL) || (be->buffer == NULL)) {
#ifdef PRINT_WARNINGS      
    print("sendBuffer called with invalid arguments (NULL).\n");
#endif
    return;
  }  
  if (be->len == 0)       
    return; /* nothing to send */  
  if (outgoingCheck(be->prio) == 0) {
    be->len = 0; /* too busy, drop! */
    return;
  }

  crc = crc32N(be->buffer, be->len);
#ifdef DUMP_PACKETS_SEND
  print("Send Buffer sends to %d:\n",
	&hex);
  dumpPacket(be->buffer, be->len);
#endif
  
  if (be->len ==
      encryptBlock(be->buffer,
		   be->len,
		   &be->skey,
		   (BLOWFISHEncryptedData*)&content)) {
    sendToHost(&be->hostId,
	       be->len,
	       &content,
	       ENCRYPTED_FLAG,
	       crc);  
  }
  be->len = 0; /* ok, mark buffer as send */
#ifdef PRINT_CONNECTION
  print("CONNECTION: done sending buffer of host %s.\n",
	&hex);
#endif
}

/* end of connection.c */
