/**
 * General key management (high level). This wraps
 * around hostkey and configuration and storage providing
 * simple access to the database of keys.
 * @author Christian Grothoff
 * @file primitives/keyservice.c
 **/

#include "keyservice.h"

/**
 * The SECRET hostkey. Keep local, never export outside of this
 * module and hostkey.c!
 **/
HOSTKEY hostkey;

/**
 * The public hostkey
 **/
PublicKey * publicKey;

/**
 * Initialize KeyService. Configuration must be 
 * initialized at this point. You must call this
 * method first!
 **/
void initKeyService() {
  FileName hostkeyfile;
  HostKeyEncoded * encHostkey;
  HELO_Message helo;
  BLOCK_LENGTH len;
  int res;

  initRAND();
  hostkeyfile = getHostKeyFilePriv();
  hostkeyfile = expandFileName(hostkeyfile);
  res = readFile(hostkeyfile, BLOCK_LENGTH_SIZE, &len);
  if (res == BLOCK_LENGTH_SIZE) {
    encHostkey = (HostKeyEncoded*) malloc(len+BLOCK_LENGTH_SIZE);    
    if (len+BLOCK_LENGTH_SIZE != 
	readFile(hostkeyfile, len+BLOCK_LENGTH_SIZE, encHostkey)) {       
      free(encHostkey);
      encHostkey = NULL;
    }
  } else
    encHostkey = NULL;
  if (encHostkey == NULL) { /* make new hostkey */
    print("Creating new hostkey (this may take a while)... ");
    hostkey = makeHostkey();
    encHostkey = encodeHostkey(hostkey);
    writeFile(hostkeyfile, 
	      encHostkey, 
	      encHostkey->len+BLOCK_LENGTH_SIZE);
    print("Done.\n");
   } else {
    hostkey = decodeHostkey(encHostkey);
  }
  publicKey = malloc(sizeof(PublicKey));
  getPublicKey(hostkey, publicKey);
  createHELO(&helo);
  bindAddress(&helo);
  free(encHostkey); /* ok on encodeHostKey path! */
  free(hostkeyfile);
}

/**
 * Get the public key of the host
 * @return reference to the public key. Do not free it!
 **/
PublicKey * getPublicHostkey() {
  return publicKey;
}

/**
 * Obtain identity from publicHostkey.
 * @param pubKey the public key of the host
 * @param result address where to write the identity of the node
 **/
void getHostIdentity(PublicKey * pubKey,
		     HostIdentity * result) {
  hash(pubKey,
       pubKey->len+BLOCK_LENGTH_SIZE,
       &result->hashPubKey);
}

/**
 * Check if 2 hosts are the same (returns 1 if yes)
 **/
int hostIdentityEquals(HostIdentity * first, 
		       HostIdentity * second) {
  if ((first == NULL) || (second == NULL))
    return 0;
  return 
    (first->hashPubKey.a == second->hashPubKey.a) &&
    (first->hashPubKey.b == second->hashPubKey.b) &&
    (first->hashPubKey.c == second->hashPubKey.c) &&
    (first->hashPubKey.d == second->hashPubKey.d) &&
    (first->hashPubKey.e == second->hashPubKey.e);
}

/** 
 * Sign arbitrary data. ALWAYS use only on data we generated
 * entirely! 
 **/
int signData(void * data,
	     BLOCK_LENGTH size,
	     Signature * result) {
  return sign(hostkey, size, data, result);
}

/**
 * Decrypt a given block with the hostkey. 
 * @param block the data to decrypt, encoded as returned by encrypt, not consumed
 * @param result pointer to a location where the result can be stored
 * @param max the maximum number of bits to store for the result, if
 *        the decrypted block is bigger, an error is returned
 * @returns the size of the decrypted block, -1 on error
 **/
int decryptData(RSAEncryptedData * block,
		void * result,
		unsigned int max) {
  return decrypt(hostkey, block, result, max);
}



/* end of keyservice */
